@@ -16,11 +16,13 @@
 import os
 import signal
 import subprocess
+import time
 
 from bzrlib import (
-    branch,
+    branch as _mod_branch,
     bzrdir,
     errors,
+    lazy_regex,
     merge,
     revision,
     revisionspec,
@@ -28,6 +30,7 @@ from bzrlib import (
     trace,
     transport,
     urlutils,
+    version_info as bzr_version_info,
     )
 
 try:
@@ -55,29 +58,312 @@ MERGE_INSTRUCTION = "merge"
 NEST_PART_INSTRUCTION = "nest-part"
 NEST_INSTRUCTION = "nest"
 RUN_INSTRUCTION = "run"
+USAGE = {
+    MERGE_INSTRUCTION: 'merge NAME BRANCH [REVISION]',
+    NEST_INSTRUCTION: 'nest NAME BRANCH TARGET-DIR [REVISION]',
+    NEST_PART_INSTRUCTION:
+        'nest-part NAME BRANCH SUBDIR [TARGET-DIR [REVISION]]',
+    RUN_INSTRUCTION: 'run COMMAND',
+    }
 
 SAFE_INSTRUCTIONS = [
     MERGE_INSTRUCTION, NEST_PART_INSTRUCTION, NEST_INSTRUCTION]
 
-TIME_VAR = "{time}"
-DATE_VAR = "{date}"
-REVNO_VAR = "{revno}"
-REVNO_PARAM_VAR = "{revno:%s}"
-DEBUPSTREAM_VAR = "{debupstream}"
 
-ok_to_preserve = [DEBUPSTREAM_VAR]
+class SubstitutionUnavailable(errors.BzrError):
+    _fmt = """Substitution for %(name)s not available: %(reason)s"""
+
+    def __init__(self, name, reason):
+        errors.BzrError.__init__(self, name=name, reason=reason)
+
+
+class SubstitutionVariable(object):
+    """A substitution variable for a version string."""
+
+    def replace(self, value):
+        """Replace name with value."""
+        raise NotImplementedError(self.replace)
+
+
+class SimpleSubstitutionVariable(SubstitutionVariable):
+
+    name = None
+
+    def replace(self, value):
+        if not self.name in value:
+            return value
+        return value.replace(self.name, self.get())
+
+    def get(self):
+        raise NotImplementedError(self.value)
+
+
+class BranchSubstitutionVariable(SimpleSubstitutionVariable):
+
+    basename = None
+
+    def __init__(self, branch_name=None):
+        super(BranchSubstitutionVariable, self).__init__()
+        self.branch_name = branch_name
+
+    @classmethod
+    def determine_name(cls, branch_name):
+        if branch_name is None:
+            return "{%s}" % cls.basename
+        else:
+            return "{%s:%s}" % (cls.basename, branch_name)
+
+    @property
+    def name(self):
+        return self.determine_name(self.branch_name)
+
+
+class TimeVariable(SimpleSubstitutionVariable):
+
+    name = "{time}"
+
+    def __init__(self, time):
+        self._time = time
+
+    def get(self):
+        return self._time.strftime("%Y%m%d%H%M")
+
+
+class DateVariable(SimpleSubstitutionVariable):
+
+    name = "{date}"
+
+    def __init__(self, time):
+        self._time = time
+
+    def get(self):
+        return self._time.strftime("%Y%m%d")
+
+
+class DebUpstreamVariable(BranchSubstitutionVariable):
+
+    basename = "debupstream"
+
+    def __init__(self, branch_name, version):
+        super(DebUpstreamVariable, self).__init__(branch_name)
+        self._version = version
+
+    @classmethod
+    def from_changelog(cls, branch_name, changelog):
+        if len(changelog._blocks) > 0:
+            return cls(branch_name, changelog._blocks[0].version)
+        else:
+            return cls(branch_name, None)
+
+    def get(self):
+        if self._version is None:
+            raise SubstitutionUnavailable(self.name,
+                "No previous changelog to take the upstream version from")
+        # Should we include the epoch?
+        return self._version.upstream_version
+
+
+class DebVersionVariable(BranchSubstitutionVariable):
+
+    basename = "debversion"
+
+    def __init__(self, branch_name, version):
+        super(DebVersionVariable, self).__init__(branch_name)
+        self._version = version
+
+    @classmethod
+    def from_changelog(cls, branch_name, changelog):
+        if len(changelog._blocks) > 0:
+            return cls(branch_name, changelog._blocks[0].version)
+        else:
+            return cls(branch_name, None)
+
+    def get(self):
+        if self._version is None:
+            raise SubstitutionUnavailable(self.name,
+                "No previous changelog to take the version from")
+        return str(self._version)
+
+
+class DebUpstreamBaseVariable(DebUpstreamVariable):
+
+    basename = "{debupstream-base}"
+    version_regex = lazy_regex.lazy_compile(r'([~+])(svn[0-9]+|bzr[0-9]+|git[0-9a-f]+)')
+
+    def get(self):
+        version = super(DebUpstreamBaseVariable, self).get()
+        version = self.version_regex.sub("\\1", version)
+        if version[-1] not in ("~", "+"):
+            version += "+"
+        return version
+
+
+class RevisionVariable(BranchSubstitutionVariable):
+
+    def __init__(self, branch_name, branch, revid):
+        super(RevisionVariable, self).__init__(branch_name)
+        self.branch = branch
+        self.revid = revid
+
+
+class RevnoVariable(RevisionVariable):
+
+    basename = "revno"
+
+    def get_revno(self):
+        try:
+            revno = self.branch.revision_id_to_revno(self.revid)
+            return str(revno)
+        except errors.NoSuchRevision:
+            # We need to load and use the full revno map after all
+            result = self.branch.get_revision_id_to_revno_map().get(
+                    self.revid)
+        if result is None:
+            return result
+        return ".".join(result)
+
+    def get(self):
+        revno = self.get_revno()
+        if revno is None:
+            raise errors.BzrCommandError("Can't substitute revno of "
+                    "branch %s in deb-version, as it's revno can't be "
+                    "determined" % revno)
+        return revno
+
+
+class RevtimeVariable(RevisionVariable):
+
+    basename = "revtime"
+
+    def get(self):
+        rev = self.branch.repository.get_revision(self.revid)
+        return time.strftime("%Y%m%d%H%M", time.gmtime(rev.timestamp))
+
+
+class RevdateVariable(RevisionVariable):
+
+    basename = "revdate"
+
+    def get(self):
+        rev = self.branch.repository.get_revision(self.revid)
+        return time.strftime("%Y%m%d", time.gmtime(rev.timestamp))
+
+
+def extract_svn_revnum(rev):
+    try:
+        foreign_revid = rev.foreign_revid
+    except AttributeError:
+        try:
+            (mapping_name, uuid, bp, srevnum) = rev.revision_id.split(":", 3)
+        except ValueError:
+            raise errors.InvalidRevisionId(rev.revision_id, None)
+        if not mapping_name.startswith("svn-"):
+            raise errors.InvalidRevisionId(rev.revision_id, None)
+        return int(srevnum)
+    else:
+        if rev.mapping.vcs.abbreviation == "svn":
+            return foreign_revid[2]
+        else:
+            raise errors.InvalidRevisionId(rev.revision_id, None)
+
+
+class SubversionRevnumVariable(RevisionVariable):
+
+    basename = "svn-revno"
+
+    def get(self):
+        rev = self.branch.repository.get_revision(self.revid)
+        try:
+            revno = extract_svn_revnum(rev)
+        except errors.InvalidRevisionId:
+            raise errors.BzrCommandError("unable to expand %s for %r in %r: "
+                "not a Subversion revision" % (
+                    self.name, self.revid, self.branch))
+        return str(revno)
+
+
+def extract_git_foreign_revid(rev):
+    try:
+        foreign_revid = rev.foreign_revid
+    except AttributeError:
+        try:
+            (mapping_name, foreign_revid) = rev.revision_id.split(":", 1)
+        except ValueError:
+            raise errors.InvalidRevisionId(rev.revision_id, None)
+        if not mapping_name.startswith("git-"):
+            raise errors.InvalidRevisionId(rev.revision_id, None)
+        return foreign_revid
+    else:
+        if rev.mapping.vcs.abbreviation == "git":
+            return foreign_revid
+        else:
+            raise errors.InvalidRevisionId(rev.revision_id, None)
+
+
+class GitCommitVariable(RevisionVariable):
+
+    basename = "git-commit"
+
+    def get(self):
+        rev = self.branch.repository.get_revision(self.revid)
+        try:
+            commit_sha = extract_git_foreign_revid(rev)
+        except errors.InvalidRevisionId:
+            raise errors.BzrCommandError("unable to expand %s for %r in %r: "
+                "not a Git revision" % (
+                    self.name, self.revid, self.branch))
+        return commit_sha[:7]
+
+
+class LatestTagVariable(RevisionVariable):
+
+    basename = "latest-tag"
+
+    def get(self):
+        reverse_tag_dict = self.branch.tags.get_reverse_tag_dict()
+        self.branch.lock_read()
+        try:
+            graph = self.branch.repository.get_graph()
+            for revid in graph.iter_lefthand_ancestry(self.revid):
+                if revid in reverse_tag_dict:
+                    return reverse_tag_dict[revid][0]
+            else:
+                raise errors.BzrCommandError("No tags set on branch %s mainline" %
+                    self.branch_name)
+        finally:
+            self.branch.unlock()
+
+
+ok_to_preserve = [DebUpstreamVariable, DebUpstreamBaseVariable,
+    DebVersionVariable]
 # The variables that don't require substitution in their name
-simple_vars = [TIME_VAR, DATE_VAR, REVNO_VAR, DEBUPSTREAM_VAR]
+simple_vars = [TimeVariable, DateVariable]
+branch_vars = [RevnoVariable, SubversionRevnumVariable,
+    GitCommitVariable, LatestTagVariable, DebVersionVariable,
+    DebUpstreamBaseVariable, DebUpstreamVariable, RevdateVariable,
+    RevtimeVariable]
 
 
 def check_expanded_deb_version(base_branch):
     checked_version = base_branch.deb_version
+    if checked_version is None:
+        return
     for token in ok_to_preserve:
-        checked_version = checked_version.replace(token, "")
+        if issubclass(token, BranchSubstitutionVariable):
+            for name in base_branch.list_branch_names():
+                checked_version = checked_version.replace(
+                    token.determine_name(name), "")
+            checked_version = checked_version.replace(
+                    token.determine_name(None), "")
+        else:
+            checked_version = checked_version.replace(
+                token.name, "")
     if "{" in checked_version:
-        available_tokens = simple_vars
-        for name in base_branch.list_branch_names():
-            available_tokens.append(REVNO_PARAM_VAR % name)
+        available_tokens = [var.name for var in simple_vars]
+        for var_kls in branch_vars:
+            for name in base_branch.list_branch_names():
+                available_tokens.append(var_kls.determine_name(name))
+            available_tokens.append(var_kls.determine_name(None))
         raise errors.BzrCommandError("deb-version not fully "
                 "expanded: %s. Valid substitutions are: %s"
                 % (base_branch.deb_version, available_tokens))
@@ -141,7 +427,8 @@ def pull_or_branch(tree_to, br_to, br_from, to_transport, revision_id,
         dir = br_from.bzrdir.sprout(to_transport.base, revision_id,
                                     possible_transports=possible_transports,
                                     accelerator_tree=accelerator_tree,
-                                    source_branch=br_from)
+                                    source_branch=br_from,
+                                    stacked=(bzr_version_info >= (2, 3, 0)))
         try:
             tree_to = dir.open_workingtree()
         except errors.NoWorkingTree:
@@ -199,7 +486,7 @@ def merge_branch(child_branch, tree_to, br_to, possible_transports=None):
     :param br_to: the Branch to merge in to.
     """
     if child_branch.branch is None:
-        child_branch.branch = branch.Branch.open(child_branch.url,
+        child_branch.branch = _mod_branch.Branch.open(child_branch.url,
                 possible_transports=possible_transports)
     child_branch.branch.lock_read()
     try:
@@ -219,14 +506,21 @@ def merge_branch(child_branch, tree_to, br_to, possible_transports=None):
         else:
             merge_revid = child_branch.branch.last_revision()
         child_branch.revid = merge_revid
-        merger = merge.Merger.from_revision_ids(None, tree_to, merge_revid,
-                other_branch=child_branch.branch, tree_branch=br_to)
+        try:
+            merger = merge.Merger.from_revision_ids(None, tree_to, merge_revid,
+                    other_branch=child_branch.branch, tree_branch=br_to)
+        except errors.UnrelatedBranches:
+            # Let's just try and hope for the best.
+            merger = merge.Merger.from_revision_ids(None, tree_to, merge_revid,
+                    other_branch=child_branch.branch, tree_branch=br_to,
+                    base=revision.NULL_REVISION)
         merger.merge_type = merge.Merge3Merger
         if (merger.base_rev_id == merger.other_rev_id and
                 merger.other_rev_id is not None):
             # Nothing to do.
             return
         conflict_count = merger.do_merge()
+
         merger.set_pending()
         if conflict_count:
             # FIXME: better reporting
@@ -255,31 +549,25 @@ def nest_part_branch(child_branch, tree_to, br_to, subpath, target_subdir=None):
     :param target_subdir: (optional) directory in target to merge that
         subpath into.  Defaults to basename of subpath.
     """
-    merge_from = branch.Branch.open(child_branch.url)
-    merge_from.lock_read()
+    child_branch.branch = _mod_branch.Branch.open(child_branch.url)
+    child_branch.branch.lock_read()
     try:
-        if child_branch.revspec is not None:
-            merge_revspec = revisionspec.RevisionSpec.from_string(
-                    child_branch.revspec)
-            merge_revid = merge_revspec.as_revision_id(merge_from)
-        else:
-            merge_revid = merge_from.last_revision()
-        child_branch.revid = merge_revid
-        other_tree = merge_from.basis_tree()
+        child_branch.resolve_revision_id()
+        other_tree = child_branch.branch.basis_tree()
         other_tree.lock_read()
         try:
             if target_subdir is None:
                 target_subdir = os.path.basename(subpath)
             merger = MergeIntoMerger(this_tree=tree_to, other_tree=other_tree,
-                other_branch=merge_from, target_subdir=target_subdir,
-                source_subpath=subpath, other_rev_id=merge_revid)
-            merger.set_base_revision(revision.NULL_REVISION, merge_from)
+                other_branch=child_branch.branch, target_subdir=target_subdir,
+                source_subpath=subpath, other_rev_id=child_branch.revid)
+            merger.set_base_revision(revision.NULL_REVISION, child_branch.branch)
             conflict_count = merger.do_merge()
             merger.set_pending()
         finally:
             other_tree.unlock()
     finally:
-        merge_from.unlock()
+        child_branch.branch.unlock()
 
     if conflict_count:
         # FIXME: better reporting
@@ -291,50 +579,27 @@ def nest_part_branch(child_branch, tree_to, br_to, subpath, target_subdir=None):
 def update_branch(base_branch, tree_to, br_to, to_transport,
         possible_transports=None):
     if base_branch.branch is None:
-        base_branch.branch = branch.Branch.open(base_branch.url,
+        base_branch.branch = _mod_branch.Branch.open(base_branch.url,
                 possible_transports=possible_transports)
     base_branch.branch.lock_read()
     try:
-        if base_branch.revspec is not None:
-            revspec = revisionspec.RevisionSpec.from_string(
-                    base_branch.revspec)
-            revision_id = revspec.as_revision_id(base_branch.branch)
-        else:
-            revision_id = base_branch.branch.last_revision()
-        base_branch.revid = revision_id
+        base_branch.resolve_revision_id()
         tree_to, br_to = pull_or_branch(tree_to, br_to, base_branch.branch,
-                to_transport, revision_id,
+                to_transport, base_branch.revid,
                 possible_transports=possible_transports)
     finally:
         base_branch.branch.unlock()
     return tree_to, br_to
 
 
-def _resolve_revisions_recurse(new_branch, substitute_revno,
+def _resolve_revisions_recurse(new_branch, substitute_branch_vars,
         if_changed_from=None):
     changed = False
-    new_branch.branch = branch.Branch.open(new_branch.url)
+    new_branch.branch = _mod_branch.Branch.open(new_branch.url)
     new_branch.branch.lock_read()
     try:
-        if new_branch.revspec is not None:
-            revspec = revisionspec.RevisionSpec.from_string(
-                    new_branch.revspec)
-            revision_id = revspec.as_revision_id(new_branch.branch)
-        else:
-            revision_id = new_branch.branch.last_revision()
-        new_branch.revid = revision_id
-        def get_revno():
-            try:
-                revno = new_branch.branch.revision_id_to_revno(revision_id)
-                return str(revno)
-            except errors.NoSuchRevision:
-                # We need to load and use the full revno map after all
-                result = new_branch.branch.get_revision_id_to_revno_map().get(
-                        revision_id)
-            if result is None:
-                return result
-            return ".".join(result)
-        substitute_revno(new_branch.name, get_revno)
+        new_branch.resolve_revision_id()
+        substitute_branch_vars(new_branch.name, new_branch.branch, new_branch.revid)
         if (if_changed_from is not None
                 and (new_branch.revspec is not None
                         or if_changed_from.revspec is not None)):
@@ -345,7 +610,7 @@ def _resolve_revisions_recurse(new_branch, substitute_revno,
                         new_branch.branch)
             else:
                 changed_revision_id = new_branch.branch.last_revision()
-            if revision_id != changed_revision_id:
+            if new_branch.revid != changed_revision_id:
                 changed = True
         for index, instruction in enumerate(new_branch.child_branches):
             child_branch = instruction.recipe_branch
@@ -354,7 +619,7 @@ def _resolve_revisions_recurse(new_branch, substitute_revno,
                 if_changed_child = if_changed_from.child_branches[index].recipe_branch
             if child_branch is not None:
                 child_changed = _resolve_revisions_recurse(child_branch,
-                        substitute_revno,
+                        substitute_branch_vars,
                         if_changed_from=if_changed_child)
                 if child_changed:
                     changed = child_changed
@@ -384,7 +649,7 @@ def resolve_revisions(base_branch, if_changed_from=None):
     if changed:
         if_changed_from_revisions = None
     changed_revisions = _resolve_revisions_recurse(base_branch,
-            base_branch.substitute_revno,
+            base_branch.substitute_branch_vars,
             if_changed_from=if_changed_from_revisions)
     if not changed:
         changed = changed_revisions
@@ -473,6 +738,9 @@ class ChildBranch(object):
             revid_part = ""
         return revid_part
 
+    def __repr__(self):
+        return "<%s %r>" % (self.__class__.__name__, self.nest_path)
+
 
 class CommandInstruction(ChildBranch):
 
@@ -506,6 +774,9 @@ class MergeInstruction(ChildBranch):
             MERGE_INSTRUCTION, self.recipe_branch.name,
             self.recipe_branch.url, revid_part)
 
+    def __repr__(self):
+        return "<%s %r>" % (self.__class__.__name__, self.recipe_branch.name)
+
 
 class NestPartInstruction(ChildBranch):
 
@@ -520,11 +791,16 @@ class NestPartInstruction(ChildBranch):
 
     def as_text(self):
         revid_part = self._get_revid_part()
-        if self.target_subdir is not None:
+        if revid_part:
+            target_subdir = self.target_subdir
+            if target_subdir is None:
+                target_subdir = self.subpath
             target_revid_part = " %s%s" % (
-                self.target_subdir, revid_part)
+                target_subdir, revid_part)
+        elif self.target_subdir is not None:
+            target_revid_part = " %s" % self.target_subdir
         else:
-            target_revid_part = revid_part
+            target_revid_part = ""
         return "%s %s %s %s%s" % (
             NEST_PART_INSTRUCTION, self.recipe_branch.name,
             self.recipe_branch.url, self.subpath, target_revid_part)
@@ -545,6 +821,10 @@ class NestInstruction(ChildBranch):
             NEST_INSTRUCTION, self.recipe_branch.name,
             self.recipe_branch.url, self.nest_path, revid_part)
 
+    def __repr__(self):
+        return "<%s %r>" % (self.__class__.__name__,
+            self.recipe_branch.name)
+
 
 class RecipeBranch(object):
     """A nested structure that represents a Recipe.
@@ -576,6 +856,16 @@ class RecipeBranch(object):
         self.revid = None
         self.branch = None
 
+    def resolve_revision_id(self):
+        """Resolve the revision id for this branch.
+        """
+        if self.revspec is not None:
+            revspec = revisionspec.RevisionSpec.from_string(self.revspec)
+            revision_id = revspec.as_revision_id(self.branch)
+        else:
+            revision_id = self.branch.last_revision()
+        self.revid = revision_id
+
     def merge_branch(self, branch):
         """Merge a child branch in to this one.
 
@@ -640,14 +930,33 @@ class RecipeBranch(object):
                 return True
         return False
 
-    def _list_child_names(self):
-        child_names = []
+    def iter_all_instructions(self):
+        """Iter over all instructions under this branch."""
+        for instruction in self.child_branches:
+            yield instruction
+            child_branch = instruction.recipe_branch
+            if child_branch is None:
+                continue
+            for instruction in child_branch.iter_all_instructions():
+                yield instruction
+
+    def iter_all_branches(self):
+        """Iterate over all branches."""
+        yield self
         for instruction in self.child_branches:
             child_branch = instruction.recipe_branch
             if child_branch is None:
                 continue
-            child_names += child_branch.list_branch_names()
-        return child_names
+            for subbranch in child_branch.iter_all_branches():
+                yield subbranch
+
+    def lookup_branch(self, name):
+        """Lookup a branch by its name."""
+        for branch in self.iter_all_branches():
+            if branch.name == name:
+                return branch
+        else:
+            raise KeyError(name)
 
     def list_branch_names(self):
         """List all of the branch names under this one.
@@ -655,7 +964,11 @@ class RecipeBranch(object):
         :return: a list of the branch names.
         :rtype: list(str)
         """
-        return [self.name] + self._list_child_names()
+        return [branch.name for branch in self.iter_all_branches()
+                if branch.name is not None]
+
+    def __repr__(self):
+        return "<%s %r>" % (self.__class__.__name__, self.name)
 
 
 class BaseRecipeBranch(RecipeBranch):
@@ -671,50 +984,65 @@ class BaseRecipeBranch(RecipeBranch):
         self.deb_version = deb_version
         self.format = format
 
-    def substitute_revno(self, branch_name, get_revno_cb):
-        """Substitute the revno for the given branch name in deb_version.
+    def substitute_branch_vars(self, branch_name, branch, revid):
+        """Substitute the branch variables for the given branch name in deb_version.
 
         Where deb_version has a place to substitute the revno for a branch
         this will substitute it for the given branch name.
 
         :param branch_name: the name of the RecipeBranch to substitute.
-        :param get_revno_cb: a callback to get the revno for that branch if
-            needed.
+        :param branch: Branch object for the branch
+        :param revid: Revision id in the branch for which to return the revno
         """
-        if branch_name is None:
-            subst_string = REVNO_VAR
-        else:
-            subst_string = REVNO_PARAM_VAR % branch_name
-        if subst_string in self.deb_version:
-            revno = get_revno_cb()
-            if revno is None:
-                raise errors.BzrCommandError("Can't substitute revno of "
-                        "branch %s in deb-version, as it's revno can't be "
-                        "determined")
-            self.deb_version = self.deb_version.replace(subst_string, revno)
+        if self.deb_version is None:
+            return
+        revno_var = RevnoVariable(branch_name, branch, revid)
+        self.deb_version = revno_var.replace(self.deb_version)
+        if self.format in (0.1, 0.2, 0.3):
+            return
+        svn_revno_var = SubversionRevnumVariable(branch_name, branch, revid)
+        self.deb_version = svn_revno_var.replace(self.deb_version)
+        git_commit_var = GitCommitVariable(branch_name, branch, revid)
+        self.deb_version = git_commit_var.replace(self.deb_version)
+        latest_tag_var = LatestTagVariable(branch_name, branch, revid)
+        self.deb_version = latest_tag_var.replace(self.deb_version)
+        revdate_var = RevdateVariable(branch_name, branch, revid)
+        self.deb_version = revdate_var.replace(self.deb_version)
+        revtime_var = RevtimeVariable(branch_name, branch, revid)
+        self.deb_version = revtime_var.replace(self.deb_version)
+        tree = branch.repository.revision_tree(revid)
+        cl_file_id = tree.path2id("debian/changelog")
+        if cl_file_id is not None:
+            tree.lock_read()
+            try:
+                cl = changelog.Changelog(tree.get_file(cl_file_id))
+                self.substitute_changelog_vars(branch_name, cl)
+            finally:
+                tree.unlock()
+
+    def substitute_changelog_vars(self, branch_name, changelog):
+        """Substitute variables related from a changelog.
+
+        :param branch_name: Branch name (None for root branch)
+        :param changelog: Changelog object to use
+        """
+        debupstream_var = DebUpstreamVariable.from_changelog(branch_name, changelog)
+        self.deb_version = debupstream_var.replace(self.deb_version)
+        debupstreambase_var = DebUpstreamBaseVariable.from_changelog(
+            branch_name, changelog)
+        self.deb_version = debupstreambase_var.replace(self.deb_version)
+        pkgversion_var = DebVersionVariable.from_changelog(branch_name, changelog)
+        self.deb_version = pkgversion_var.replace(self.deb_version)
 
     def substitute_time(self, time):
         """Substitute the time in to deb_version if needed.
 
         :param time: a datetime.datetime with the desired time.
         """
-        if TIME_VAR in self.deb_version:
-            self.deb_version = self.deb_version.replace(TIME_VAR,
-                    time.strftime("%Y%m%d%H%M"))
-        if DATE_VAR in self.deb_version:
-            self.deb_version = self.deb_version.replace(DATE_VAR,
-                    time.strftime("%Y%m%d"))
-
-    def substitute_debupstream(self, version):
-        """Substitute {debupstream} in to deb_version if needed.
-
-        :param version: the Version object to take the upstream version
-            from.
-        """
-        if DEBUPSTREAM_VAR in self.deb_version:
-            # Should we include the epoch?
-            self.deb_version = self.deb_version.replace(DEBUPSTREAM_VAR,
-                    version.upstream_version)
+        if self.deb_version is None:
+            return
+        self.deb_version = TimeVariable(time).replace(self.deb_version)
+        self.deb_version = DateVariable(time).replace(self.deb_version)
 
     def _add_child_branches_to_manifest(self, child_branches, indent_level):
         manifest = ""
@@ -727,17 +1055,15 @@ class BaseRecipeBranch(RecipeBranch):
                     indent_level+1)
         return manifest
 
-
     def __str__(self):
         return self.get_recipe_text(validate=True)
 
-    def list_branch_names(self):
-        return self._list_child_names()
-
     def get_recipe_text(self, validate=False):
-        manifest = "# bzr-builder format %s deb-version " % str(self.format)
-        # TODO: should we store the expanded version that was used?
-        manifest += "%s\n" % (self.deb_version,)
+        manifest = "# bzr-builder format %s" % str(self.format)
+        if self.deb_version is not None:
+            # TODO: should we store the expanded version that was used?
+            manifest += " deb-version %s" % (self.deb_version,)
+        manifest += "\n"
         if self.revid is not None:
             manifest += "%s revid:%s\n" % (self.url, self.revid)
         elif self.revspec is not None:
@@ -762,6 +1088,14 @@ class RecipeParseError(errors.BzrError):
                 problem=problem)
 
 
+class InstructionParseError(RecipeParseError):
+    _fmt = RecipeParseError._fmt + "\nUsage: %(usage)s"
+
+    def __init__(self, filename, line, char, problem, instruction):
+        RecipeParseError.__init__(self, filename, line, char, problem)
+        self.usage = USAGE[instruction]
+
+
 class ForbiddenInstructionError(RecipeParseError):
 
     def __init__(self, filename, line, char, problem, instruction_name=None):
@@ -779,7 +1113,7 @@ class RecipeParser(object):
     eol_char = "\n"
     digit_chars = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9")
 
-    NEWEST_VERSION = 0.3
+    NEWEST_VERSION = 0.4
 
     def __init__(self, f, filename=None):
         """Create a RecipeParser.
@@ -848,17 +1182,18 @@ class RecipeParser(object):
                 instruction = self.parse_instruction(
                     permitted_instructions=permitted_instructions)
                 if instruction == RUN_INSTRUCTION:
-                    self.parse_whitespace("the command")
+                    self.parse_whitespace("the command",
+                        instruction=instruction)
                     command = self.take_to_newline().strip()
                     self.new_line()
                     active_branches[-1].run_command(command)
                 else:
-                    branch_id = self.parse_branch_id()
-                    url = self.parse_branch_url()
+                    branch_id = self.parse_branch_id(instruction)
+                    url = self.parse_branch_url(instruction)
                     if instruction == NEST_INSTRUCTION:
-                        location = self.parse_branch_location()
+                        location = self.parse_branch_location(instruction)
                     if instruction == NEST_PART_INSTRUCTION:
-                        path = self.parse_subpath()
+                        path = self.parse_subpath(instruction)
                         target_subdir = self.parse_optional_path()
                         if target_subdir == '':
                             target_subdir = None
@@ -867,7 +1202,7 @@ class RecipeParser(object):
                             revspec = self.parse_optional_revspec()
                     else:
                         revspec = self.parse_optional_revspec()
-                    self.new_line()
+                    self.new_line(instruction)
                     last_branch = RecipeBranch(branch_id, url, revspec=revspec)
                     if instruction == NEST_INSTRUCTION:
                         active_branches[-1].nest_branch(location, last_branch)
@@ -889,9 +1224,7 @@ class RecipeParser(object):
         if version > self.NEWEST_VERSION:
             self.throw_parse_error("Unknown format: '%s'" % str(version))
         self.take_chars(len(version_str))
-        self.parse_word("deb-version")
-        self.parse_whitespace("a value for 'deb-version'")
-        deb_version = self.take_to_whitespace("a value for 'deb-version'")
+        deb_version = self.parse_optional_deb_version()
         self.new_line()
         return version, deb_version
 
@@ -922,12 +1255,13 @@ class RecipeParser(object):
         self.throw_parse_error("Expecting %s, got '%s'"
                 % (options_str, instruction))
 
-    def parse_branch_id(self):
-        self.parse_whitespace("the branch id")
+    def parse_branch_id(self, instruction):
+        self.parse_whitespace("the branch id", instruction=instruction)
         branch_id = self.peek_to_whitespace()
         if branch_id is None:
             self.throw_parse_error("End of line while looking for the "
-                    "branch id")
+                    "branch id", cls=InstructionParseError,
+                    instruction=instruction)
         if branch_id in self.seen_nicks:
             self.throw_parse_error("'%s' was already used to identify "
                     "a branch." % branch_id)
@@ -935,36 +1269,39 @@ class RecipeParser(object):
         self.seen_nicks.add(branch_id)
         return branch_id
 
-    def parse_branch_url(self):
-        self.parse_whitespace("the branch url")
-        branch_url = self.take_to_whitespace("the branch url")
+    def parse_branch_url(self, instruction):
+        self.parse_whitespace("the branch url", instruction=instruction)
+        branch_url = self.take_to_whitespace("the branch url", instruction)
         return branch_url
 
-    def parse_branch_location(self):
+    def parse_branch_location(self, instruction):
         # FIXME: Needs a better term
         self.parse_whitespace("the location to nest")
         location = self.peek_to_whitespace()
         if location is None:
             self.throw_parse_error("End of line while looking for the "
-                    "location to nest")
+                    "location to nest", cls=InstructionParseError,
+                    instruction=instruction)
         norm_location = os.path.normpath(location)
         if norm_location in self.seen_paths:
             self.throw_parse_error("The path '%s' is a duplicate of "
                     "the one used on line %d." % (location,
-                        self.seen_paths[norm_location]))
+                        self.seen_paths[norm_location]),
+                    InstructionParseError, instruction=instruction)
         if os.path.isabs(norm_location):
             self.throw_parse_error("Absolute paths are not allowed: %s"
-                    % location)
+                    % location, InstructionParseError, instruction=instruction)
         if norm_location.startswith(".."):
             self.throw_parse_error("Paths outside the current directory "
-                    "are not allowed: %s" % location)
+                    "are not allowed: %s" % location,
+                    cls=InstructionParseError, instruction=instruction)
         self.take_chars(len(location))
         self.seen_paths[norm_location] = self.line_index + 1
         return location
 
-    def parse_subpath(self):
-        self.parse_whitespace("the subpath to merge")
-        location = self.take_to_whitespace("the subpath to merge")
+    def parse_subpath(self, instruction):
+        self.parse_whitespace("the subpath to merge", instruction=instruction)
+        location = self.take_to_whitespace("the subpath to merge", instruction)
         return location
 
     def parse_revspec(self):
@@ -972,6 +1309,18 @@ class RecipeParser(object):
         revspec = self.take_to_whitespace("the revspec")
         return revspec
 
+    def parse_optional_deb_version(self):
+        self.parse_whitespace("'deb-version'", require=False)
+        actual = self.peek_to_whitespace()
+        if actual is None:
+            # End of line, no version
+            return None
+        if actual != "deb-version":
+            self.throw_expecting_error("deb-version", actual)
+        self.take_chars(len("deb-version"))
+        self.parse_whitespace("a value for 'deb-version'")
+        return self.take_to_whitespace("a value for 'deb-version'")
+
     def parse_optional_revspec(self):
         self.parse_whitespace(None, require=False)
         revspec = self.peek_to_whitespace()
@@ -999,13 +1348,18 @@ class RecipeParser(object):
     def throw_eol(self, expected):
         self.throw_parse_error("End of line while looking for '%s'" % expected)
 
-    def new_line(self):
+    def new_line(self, instruction=None):
         # Jump over any whitespace
         self.parse_whitespace(None, require=False)
         remaining = self.peek_to_whitespace()
         if remaining != None:
+            kwargs = {}
+            if instruction is not None:
+                kwargs = {
+                    'cls': InstructionParseError,
+                    'instruction': instruction}
             self.throw_parse_error("Expecting the end of the line, got '%s'"
-                    % remaining)
+                    % remaining, **kwargs)
         self.index = 0
         self.line_index += 1
         if self.line_index >= len(self.lines):
@@ -1071,12 +1425,18 @@ class RecipeParser(object):
            return old_indent_level
         return None
 
-    def parse_whitespace(self, looking_for, require=True):
+    def parse_whitespace(self, looking_for, require=True, instruction=None):
         if require:
             actual = self.peek_char()
             if actual is None:
+                kwargs = {}
+                if instruction is not None:
+                    kwargs = {
+                        'cls': InstructionParseError,
+                        'instruction': instruction,
+                        }
                 self.throw_parse_error("End of line while looking for "
-                        "%s" % looking_for)
+                        "%s" % looking_for, **kwargs)
             if actual not in self.whitespace_chars:
                 self.throw_parse_error("Expecting whitespace before %s, "
                         "got '%s'." % (looking_for, actual))
@@ -1123,11 +1483,16 @@ class RecipeParser(object):
             char = self.peek_char(skip=count)
         return ret
 
-    def take_to_whitespace(self, looking_for):
+    def take_to_whitespace(self, looking_for, instruction=None):
         text = self.peek_to_whitespace()
         if text is None:
+            kwargs = {}
+            if instruction is not None:
+                kwargs = {
+                    'cls': InstructionParseError,
+                    'instruction': instruction}
             self.throw_parse_error("End of line while looking for %s"
-                    % looking_for)
+                    % looking_for, **kwargs)
         self.take_chars(len(text))
         return text
 
