@@ -15,6 +15,7 @@
 
 import datetime
 import os
+import textwrap
 
 from bzrlib import (
         errors,
@@ -22,22 +23,40 @@ from bzrlib import (
         workingtree,
         )
 from bzrlib.tests import (
+        TestCase,
         TestCaseInTempDir,
         TestCaseWithTransport,
         )
 from bzrlib.plugins.builder.recipe import (
         BaseRecipeBranch,
         build_tree,
+        DebUpstreamBaseVariable,
+        DebUpstreamVariable,
         ensure_basedir,
+        InstructionParseError,
         ForbiddenInstructionError,
+        MERGE_INSTRUCTION,
+        NEST_INSTRUCTION,
+        NEST_PART_INSTRUCTION,
+        DebVersionVariable,
         pull_or_branch,
         RecipeParser,
         RecipeBranch,
         RecipeParseError,
         resolve_revisions,
+        RUN_INSTRUCTION,
         SAFE_INSTRUCTIONS,
+        SubstitutionUnavailable,
+        USAGE,
         )
 
+try:
+    from debian import changelog
+except ImportError:
+    # In older versions of python-debian the main package was named 
+    # debian_bundle
+    from debian_bundle import changelog
+
 
 class RecipeParserTests(TestCaseInTempDir):
 
@@ -59,6 +78,13 @@ class RecipeParserTests(TestCaseInTempDir):
         self.assertEqual("recipe", exc.filename)
         return exc
 
+    def assertInstructionParseError(self, line, char, problem, instruction,
+            callable, *args, **kwargs):
+        exc = self.assertParseError(line, char, problem, callable, *args,
+                **kwargs)
+        self.assertIsInstance(exc, InstructionParseError),
+        self.assertEqual(USAGE[instruction], exc.usage)
+
     def check_recipe_branch(self, branch, name, url, revspec=None,
             num_child_branches=0, revid=None):
         self.assertEqual(name, branch.name)
@@ -76,6 +102,11 @@ class RecipeParserTests(TestCaseInTempDir):
     def test_parses_most_basic(self):
         self.get_recipe(self.basic_header_and_branch)
 
+    def test_parses_missing_deb_version(self):
+        header = "# bzr-builder format 0.3\n"
+        branch = "http://foo.org/\n"
+        self.get_recipe(header + branch)
+
     def tests_rejects_non_comment_to_start(self):
         self.assertParseError(1, 1, "Expecting '#', got 'b'",
                 self.get_recipe, "bzr-builder")
@@ -138,53 +169,69 @@ class RecipeParserTests(TestCaseInTempDir):
                 self.basic_header + "http://foo.org/\n" + "cat")
 
     def test_rejects_merge_no_name(self):
-        self.assertParseError(3, 7, "End of line while looking for "
-                "the branch id", self.get_recipe,
+        self.assertInstructionParseError(3, 7, "End of line while looking for "
+                "the branch id", MERGE_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "merge ")
 
+    def test_rejects_merge_no_name_no_space(self):
+        self.assertInstructionParseError(3, 6, "End of line while looking for "
+                "the branch id", MERGE_INSTRUCTION, self.get_recipe,
+                self.basic_header_and_branch + "merge")
+
     def test_rejects_merge_no_url(self):
-        self.assertParseError(3, 11, "End of line while looking for "
-                "the branch url", self.get_recipe,
+        self.assertInstructionParseError(3, 11, "End of line while looking for"
+                " the branch url", MERGE_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "merge foo ")
 
     def test_rejects_text_at_end_of_merge_line(self):
-        self.assertParseError(3, 17, "Expecting the end of the line, "
-                "got 'bar'", self.get_recipe,
+        self.assertInstructionParseError(3, 17,
+                "Expecting the end of the line, got 'bar'",
+                MERGE_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "merge foo url 2 bar")
 
     def test_rejects_nest_part_no_name(self):
-        self.assertParseError(3, 11, "End of line while looking for "
-                "the branch id", self.get_recipe,
+        self.assertInstructionParseError(3, 11, "End of line while looking for"
+                " the branch id", NEST_PART_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest-part ")
 
     def test_rejects_nest_part_no_url(self):
-        self.assertParseError(3, 15, "End of line while looking for "
-                "the branch url", self.get_recipe,
+        self.assertInstructionParseError(3, 15, "End of line while looking for"
+                " the branch url", NEST_PART_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest-part foo ")
 
     def test_rejects_nest_part_no_subpath(self):
-        self.assertParseError(3, 22, "End of line while looking for "
-                "the subpath to merge", self.get_recipe,
+        self.assertInstructionParseError(3, 22, "End of line while looking for"
+                " the subpath to merge", NEST_PART_INSTRUCTION,
+                self.get_recipe,
                 self.basic_header_and_branch + "nest-part foo url:// ")
 
+    def test_rejects_nest_part_no_subpath_no_space(self):
+        self.assertInstructionParseError(3, 21, "End of line while looking for"
+                " the subpath to merge", NEST_PART_INSTRUCTION,
+                self.get_recipe,
+                self.basic_header_and_branch + "nest-part foo url://")
+
     def test_rejects_nest_no_name(self):
-        self.assertParseError(3, 6, "End of line while looking for "
-                "the branch id", self.get_recipe,
+        self.assertInstructionParseError(3, 6, "End of line while looking for"
+                " the branch id", NEST_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest ")
 
     def test_rejects_nest_no_url(self):
-        self.assertParseError(3, 10, "End of line while looking for "
-                "the branch url", self.get_recipe,
-                self.basic_header_and_branch + "nest foo ")
+        self.assertInstructionParseError(3, 10,
+            "End of line while looking for the branch url",
+            NEST_INSTRUCTION, self.get_recipe,
+            self.basic_header_and_branch + "nest foo ")
 
     def test_rejects_nest_no_location(self):
-        self.assertParseError(3, 14, "End of line while looking for "
-                "the location to nest", self.get_recipe,
+        self.assertInstructionParseError(3, 14,
+                "End of line while looking for the location to nest",
+                NEST_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest foo url ")
 
     def test_rejects_text_at_end_of_nest_line(self):
-        self.assertParseError(3, 20, "Expecting the end of the line, "
-                "got 'baz'", self.get_recipe,
+        self.assertInstructionParseError(3, 20,
+                "Expecting the end of the line, got 'baz'",
+                NEST_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest foo url bar 2 baz")
 
     def test_rejects_indent_after_first_branch(self):
@@ -226,6 +273,12 @@ class RecipeParserTests(TestCaseInTempDir):
                 self.basic_header_and_branch + "merge foo url\n"
                 + "merge foo other-url\n")
 
+    def test_rejects_run_with_no_instruction(self):
+        self.assertInstructionParseError(3, 4,
+                "End of line while looking for the command",
+                RUN_INSTRUCTION, self.get_recipe,
+                self.basic_header_and_branch + "run\n")
+
     def test_builds_simplest_recipe(self):
         base_branch = self.get_recipe(self.basic_header_and_branch)
         self.check_base_recipe_branch(base_branch, "http://foo.org/")
@@ -443,45 +496,56 @@ class RecipeParserTests(TestCaseInTempDir):
         self.assertEqual("run", exc.instruction_name)
 
     def test_error_on_duplicate_path(self):
-        exc = self.assertParseError(3, 15, "The path '.' is a duplicate "
-                "of the one used on line 1.", self.get_recipe,
+        exc = self.assertInstructionParseError(3, 15,
+                "The path '.' is a duplicate of the one used on line 1.",
+                NEST_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest nest url .\n")
 
     def test_error_on_duplicate_path_with_another_nest(self):
-        exc = self.assertParseError(4, 16, "The path 'foo' is a duplicate "
-                "of the one used on line 3.", self.get_recipe,
+        exc = self.assertInstructionParseError(4, 16,
+                "The path 'foo' is a duplicate of the one used on line 3.",
+                NEST_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest nest url foo\n"
                 + "nest nest2 url foo\n")
 
     def test_duplicate_path_check_uses_normpath(self):
-        exc = self.assertParseError(3, 15, "The path 'foo/..' is a duplicate "
-                "of the one used on line 1.", self.get_recipe,
+        exc = self.assertInstructionParseError(3, 15,
+                "The path 'foo/..' is a duplicate of the one used on line 1.",
+                NEST_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest nest url foo/..\n")
 
     def test_error_absolute_path(self):
-        exc = self.assertParseError(3, 15, "Absolute paths are not allowed: "
-                "/etc/passwd", self.get_recipe, self.basic_header_and_branch
-                + "nest nest url /etc/passwd\n")
+        exc = self.assertInstructionParseError(3, 15,
+                "Absolute paths are not allowed: /etc/passwd",
+                NEST_INSTRUCTION, self.get_recipe,
+                self.basic_header_and_branch + "nest nest url /etc/passwd\n")
 
     def test_error_simple_parent_dir(self):
-        exc = self.assertParseError(3, 15, "Paths outside the current "
-                "directory are not allowed: ../foo", self.get_recipe,
+        exc = self.assertInstructionParseError(3, 15,
+                "Paths outside the current directory are not allowed: ../foo",
+                NEST_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest nest url ../foo\n")
 
     def test_error_complex_parent_dir(self):
-        exc = self.assertParseError(3, 15, "Paths outside the current "
-                "directory are not allowed: ./foo/../..", self.get_recipe,
+        exc = self.assertInstructionParseError(3, 15,
+                "Paths outside the current directory are not allowed:"
+                " ./foo/../..", NEST_INSTRUCTION, self.get_recipe,
                 self.basic_header_and_branch + "nest nest url ./foo/../..\n")
 
 
 class BuildTreeTests(TestCaseWithTransport):
 
+    if not getattr(TestCaseWithTransport, "assertPathExists", None):
+        # Compatibility with bzr < 2.4
+        def assertPathExists(self, path):
+            self.failUnlessExists(path)
+
     def test_ensure_basedir(self):
         to_transport = transport.get_transport("a")
         ensure_basedir(to_transport)
-        self.failUnlessExists("a")
+        self.assertPathExists("a")
         ensure_basedir(to_transport)
-        self.failUnlessExists("a")
+        self.assertPathExists("a")
         e = self.assertRaises(errors.BzrCommandError, ensure_basedir,
                 transport.get_transport("b/c"))
         self.assertTrue('Parent of "' in str(e))
@@ -492,7 +556,7 @@ class BuildTreeTests(TestCaseWithTransport):
         revid = source.commit("one")
         base_branch = BaseRecipeBranch("source", "1", 0.2)
         build_tree(base_branch, "target")
-        self.failUnlessExists("target")
+        self.assertPathExists("target")
         tree = workingtree.WorkingTree.open("target")
         self.assertEqual(revid, tree.last_revision())
         self.assertEqual(revid, base_branch.revid)
@@ -504,7 +568,7 @@ class BuildTreeTests(TestCaseWithTransport):
         os.mkdir("target")
         base_branch = BaseRecipeBranch("source", "1", 0.2)
         build_tree(base_branch, "target")
-        self.failUnlessExists("target")
+        self.assertPathExists("target")
         tree = workingtree.WorkingTree.open("target")
         self.assertEqual(revid, tree.last_revision())
         self.assertEqual(revid, base_branch.revid)
@@ -515,7 +579,7 @@ class BuildTreeTests(TestCaseWithTransport):
         self.make_branch_and_tree("target")
         base_branch = BaseRecipeBranch("source", "1", 0.2)
         build_tree(base_branch, "target")
-        self.failUnlessExists("target")
+        self.assertPathExists("target")
         tree = workingtree.WorkingTree.open("target")
         self.assertEqual(revid, tree.last_revision())
         self.assertEqual(revid, base_branch.revid)
@@ -535,7 +599,7 @@ class BuildTreeTests(TestCaseWithTransport):
         nested_branch = RecipeBranch("nested", "source2")
         base_branch.nest_branch("sub", nested_branch)
         build_tree(base_branch, "target")
-        self.failUnlessExists("target")
+        self.assertPathExists("target")
         tree = workingtree.WorkingTree.open("target")
         self.assertEqual([source1_rev_id], tree.get_parent_ids())
         tree = workingtree.WorkingTree.open("target/sub")
@@ -553,7 +617,7 @@ class BuildTreeTests(TestCaseWithTransport):
         merged_branch = RecipeBranch("merged", "source2")
         base_branch.merge_branch(merged_branch)
         build_tree(base_branch, "target")
-        self.failUnlessExists("target")
+        self.assertPathExists("target")
         tree = workingtree.WorkingTree.open("target")
         last_revid = tree.last_revision()
         last_revtree = tree.branch.repository.revision_tree(last_revid)
@@ -563,6 +627,33 @@ class BuildTreeTests(TestCaseWithTransport):
         self.assertEqual(source1_rev_id, base_branch.revid)
         self.assertEqual(source2_rev_id, merged_branch.revid)
 
+    def test_build_tree_merge_unrelated(self):
+        """Make a branch with one file and one commit."""
+        source1 = self.make_branch_and_tree("source1")
+        self.build_tree_contents([("source1/a", "file a")])
+        source1.add(["a"])
+        source1_rev_id = source1.commit("one")
+
+        source2 = self.make_branch_and_tree("source2")
+        self.build_tree_contents([("source2/b", "file b")])
+        source2.add(["b"])
+        source2_rev_id = source2.commit("one")
+
+        base_branch = BaseRecipeBranch("source1", "1", 0.2)
+        merged_branch = RecipeBranch("merged", "source2")
+        base_branch.merge_branch(merged_branch)
+        build_tree(base_branch, "target")
+        self.assertPathExists("target")
+        tree = workingtree.WorkingTree.open("target")
+        last_revid = tree.last_revision()
+        last_revtree = tree.branch.repository.revision_tree(last_revid)
+        self.assertEqual([source1_rev_id, source2_rev_id],
+                last_revtree.get_parent_ids())
+        self.check_file_contents("target/a", "file a")
+        self.check_file_contents("target/b", "file b")
+        self.assertEqual(source1_rev_id, base_branch.revid)
+        self.assertEqual(source2_rev_id, merged_branch.revid)
+
     def test_build_tree_nest_part(self):
         """A recipe can specify a merge of just part of an unrelated tree."""
         source1 = self.make_source_branch("source1")
@@ -629,7 +720,7 @@ class BuildTreeTests(TestCaseWithTransport):
         merged_branch2 = RecipeBranch("merged2", "source3")
         base_branch.merge_branch(merged_branch2)
         build_tree(base_branch, "target")
-        self.failUnlessExists("target")
+        self.assertPathExists("target")
         tree = workingtree.WorkingTree.open("target")
         last_revid = tree.last_revision()
         previous_revid = tree.branch.revision_history()[-2]
@@ -656,7 +747,7 @@ class BuildTreeTests(TestCaseWithTransport):
         base_branch.merge_branch(merged_branch)
         self.assertRaises(errors.BzrCommandError, build_tree,
                 base_branch, "target")
-        self.failUnlessExists("target")
+        self.assertPathExists("target")
         tree = workingtree.WorkingTree.open("target")
         self.assertEqual(source1_rev_id, tree.last_revision())
         self.assertEqual([source1_rev_id, source2_rev_id],
@@ -684,7 +775,7 @@ class BuildTreeTests(TestCaseWithTransport):
         merged_branch = RecipeBranch("merged", "source2", revspec="2")
         base_branch.merge_branch(merged_branch)
         build_tree(base_branch, "target")
-        self.failUnlessExists("target")
+        self.assertPathExists("target")
         tree = workingtree.WorkingTree.open("target")
         last_revid = tree.last_revision()
         last_revtree = tree.branch.repository.revision_tree(last_revid)
@@ -820,8 +911,8 @@ class BuildTreeTests(TestCaseWithTransport):
         base_branch = BaseRecipeBranch("source", "1", 0.2)
         base_branch.run_command("touch test")
         build_tree(base_branch, "target")
-        self.failUnlessExists("target")
-        self.failUnlessExists("target/test")
+        self.assertPathExists("target")
+        self.assertPathExists("target/test")
         tree = workingtree.WorkingTree.open("target")
         self.assertEqual(revid, tree.last_revision())
         self.assertEqual(revid, base_branch.revid)
@@ -973,9 +1064,91 @@ class ResolveRevisionsTests(TestCaseWithTransport):
         branch1 = BaseRecipeBranch("source", "{revno:packaging}", 0.2)
         self.assertRaises(errors.BzrCommandError, resolve_revisions, branch1)
 
+    def test_substitute_svn_not_svn(self):
+        br = self.make_branch("source")
+        source = br.create_checkout("checkout")
+        source.commit("one")
+        source.commit("two")
+        branch1 = BaseRecipeBranch("source", "foo-{svn-revno}", 0.4)
+        e = self.assertRaises(errors.BzrCommandError, resolve_revisions,
+            branch1)
+        self.assertTrue(str(e).startswith("unable to expand {svn-revno} "),
+            e)
+
+    def test_substitute_svn(self):
+        br = self.make_branch("source")
+        source = br.create_checkout("checkout")
+        source.commit("one")
+        source.commit("two",
+            rev_id="svn-v4:be7e6eca-30d4-0310-a8e5-ac0d63af7070:trunk:5344")
+        branch1 = BaseRecipeBranch("source", "foo-{svn-revno}", 0.4)
+        resolve_revisions(branch1)
+        self.assertEqual("foo-5344", branch1.deb_version)
+
+    def test_substitute_git_not_git(self):
+        source = self.make_branch_and_tree("source")
+        source.commit("one")
+        source.commit("two")
+        branch1 = BaseRecipeBranch("source", "foo-{git-commit}", 0.4)
+        e = self.assertRaises(errors.BzrCommandError, resolve_revisions,
+            branch1)
+        self.assertTrue(str(e).startswith("unable to expand {git-commit} "),
+            e)
+
+    def test_substitute_git(self):
+        source = self.make_branch_and_tree("source")
+        source.commit("one", 
+            rev_id="git-v1:a029d7b2cc83c26a53d8b2a24fa12c340fcfac58")
+        branch1 = BaseRecipeBranch("source", "foo-{git-commit}", 0.4)
+        resolve_revisions(branch1)
+        self.assertEqual("foo-a029d7b", branch1.deb_version)
+
+    def test_latest_tag(self):
+        source = self.make_branch_and_tree("source")
+        revid = source.commit("one")
+        source.branch.tags.set_tag("millbank", revid)
+        source.commit("two")
+        branch1 = BaseRecipeBranch("source", "foo-{latest-tag}", 0.4)
+        resolve_revisions(branch1)
+        self.assertEqual("foo-millbank", branch1.deb_version)
+
+    def test_latest_tag_no_tag(self):
+        source = self.make_branch_and_tree("source")
+        revid = source.commit("one")
+        source.commit("two")
+        branch1 = BaseRecipeBranch("source", "foo-{latest-tag}", 0.4)
+        e = self.assertRaises(errors.BzrCommandError, resolve_revisions, branch1)
+        self.assertTrue(str(e).startswith("No tags set on branch None mainline"),
+            e)
+
+    def test_substitute_revdate(self):
+        br = self.make_branch("source")
+        source = br.create_checkout("checkout")
+        source.commit("one")
+        source.commit("two", timestamp=1307708628, timezone=0)
+        branch1 = BaseRecipeBranch("source", "foo-{revdate}", 0.4)
+        resolve_revisions(branch1)
+        self.assertEqual("foo-20110610", branch1.deb_version)
+
+    def test_substitute_revtime(self):
+        br = self.make_branch("source")
+        source = br.create_checkout("checkout")
+        source.commit("one")
+        source.commit("two", timestamp=1307708628, timezone=0)
+        branch1 = BaseRecipeBranch("source", "foo-{revtime}", 0.4)
+        resolve_revisions(branch1)
+        self.assertEqual("foo-201106101223", branch1.deb_version)
+
 
 class StringifyTests(TestCaseInTempDir):
 
+    def test_missing_debversion(self):
+        base_branch = BaseRecipeBranch("base_url", None, 0.1)
+        base_branch.revid = "base_revid"
+        manifest = str(base_branch)
+        self.assertEqual("# bzr-builder format 0.1\n"
+                "base_url revid:base_revid\n", manifest)
+
     def test_simple_manifest(self):
         base_branch = BaseRecipeBranch("base_url", "1", 0.1)
         base_branch.revid = "base_revid"
@@ -1064,8 +1237,8 @@ class StringifyTests(TestCaseInTempDir):
         nested_branch1 = RecipeBranch("nested1", "nested1_url",
                 revspec="tag:foo")
         base_branch.nest_branch(".", nested_branch1)
-        self.assertRaises(RecipeParseError, base_branch.get_recipe_text,
-                validate=True)
+        self.assertRaises(InstructionParseError,
+            base_branch.get_recipe_text, validate=True)
 
     def test_str_validates(self):
         base_branch = BaseRecipeBranch("base_url", "1", 0.1)
@@ -1073,7 +1246,7 @@ class StringifyTests(TestCaseInTempDir):
         nested_branch1 = RecipeBranch("nested1", "nested1_url",
                 revspec="tag:foo")
         base_branch.nest_branch(".", nested_branch1)
-        self.assertRaises(RecipeParseError, str, base_branch)
+        self.assertRaises(InstructionParseError, str, base_branch)
 
     def test_with_nest_part(self):
         base_branch = BaseRecipeBranch("base_url", "1", 0.1)
@@ -1087,8 +1260,31 @@ class StringifyTests(TestCaseInTempDir):
                 "nest-part nested1 nested1_url foo bar tag:foo\n",
                 manifest)
 
+    def test_with_nest_part_with_no_target_dir(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1)
+        base_branch.revid = "base_revid"
+        nested_branch1 = RecipeBranch("nested1", "nested1_url",
+                revspec="tag:foo")
+        base_branch.nest_part_branch(nested_branch1, "foo", None)
+        manifest = base_branch.get_recipe_text()
+        self.assertEqual("# bzr-builder format 0.1 deb-version 1\n"
+                "base_url revid:base_revid\n"
+                "nest-part nested1 nested1_url foo foo tag:foo\n",
+                manifest)
 
-class RecipeBranchTests(TestCaseInTempDir):
+    def test_with_nest_part_with_no_target_dir_no_revspec(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1)
+        base_branch.revid = "base_revid"
+        nested_branch1 = RecipeBranch("nested1", "nested1_url")
+        base_branch.nest_part_branch(nested_branch1, "foo", None)
+        manifest = base_branch.get_recipe_text()
+        self.assertEqual("# bzr-builder format 0.1 deb-version 1\n"
+                "base_url revid:base_revid\n"
+                "nest-part nested1 nested1_url foo\n",
+                manifest)
+
+
+class RecipeBranchTests(TestCaseWithTransport):
 
     def test_base_recipe_branch(self):
         base_branch = BaseRecipeBranch("base_url", "1", 0.2, revspec="2")
@@ -1137,27 +1333,60 @@ class RecipeBranchTests(TestCaseInTempDir):
         base_branch.substitute_time(time)
         self.assertEqual("1-19700101", base_branch.deb_version)
 
-    def test_substitute_revno(self):
+    def test_substitute_branch_vars(self):
         base_branch = BaseRecipeBranch("base_url", "1", 0.2)
-        base_branch.substitute_revno(None, None)
+        wt = self.make_branch_and_tree("br")
+        revid = wt.commit("acommit")
+        base_branch.substitute_branch_vars(None, wt.branch, revid)
         self.assertEqual("1", base_branch.deb_version)
-        base_branch.substitute_revno(None, None)
+        base_branch.substitute_branch_vars(None, wt.branch, revid)
         self.assertEqual("1", base_branch.deb_version)
         base_branch = BaseRecipeBranch("base_url", "{revno}", 0.2)
-        base_branch.substitute_revno(None, lambda: "2")
-        self.assertEqual("2", base_branch.deb_version)
-        base_branch.substitute_revno(None, lambda: "2")
-        self.assertEqual("2", base_branch.deb_version)
+        base_branch.substitute_branch_vars(None, wt.branch, revid)
+        self.assertEqual("1", base_branch.deb_version)
         base_branch = BaseRecipeBranch("base_url", "{revno}", 0.2)
-        base_branch.substitute_revno("foo", None)
+        base_branch.substitute_branch_vars("foo", wt.branch, revid)
         self.assertEqual("{revno}", base_branch.deb_version)
-        base_branch.substitute_revno("foo", None)
+        base_branch.substitute_branch_vars("foo", wt.branch, revid)
         self.assertEqual("{revno}", base_branch.deb_version)
         base_branch = BaseRecipeBranch("base_url", "{revno:foo}", 0.2)
-        base_branch.substitute_revno("foo", lambda: "3")
-        self.assertEqual("3", base_branch.deb_version)
-        base_branch.substitute_revno("foo", lambda: "3")
-        self.assertEqual("3", base_branch.deb_version)
+        base_branch.substitute_branch_vars("foo", wt.branch, revid)
+        self.assertEqual("1", base_branch.deb_version)
+
+    def test_substitute_branch_vars_debupstream(self):
+        wt = self.make_branch_and_tree("br")
+        revid1 = wt.commit("acommit")
+        cl_contents = ("package (0.1-1) unstable; urgency=low\n  * foo\n"
+                    " -- maint <maint@maint.org>  Tue, 04 Aug 2009 "
+                    "10:03:10 +0100\n")
+        self.build_tree_contents(
+            [("br/debian/", ), ('br/debian/changelog', cl_contents)])
+        wt.add(['debian', 'debian/changelog'])
+        revid2 = wt.commit("with changelog")
+        base_branch = BaseRecipeBranch("base_url", "{debupstream}", 0.4)
+        # No changelog file, so no substitution
+        base_branch.substitute_branch_vars(None, wt.branch, revid1)
+        self.assertEqual("{debupstream}", base_branch.deb_version)
+        base_branch.substitute_branch_vars(None, wt.branch, revid2)
+        self.assertEqual("0.1", base_branch.deb_version)
+        base_branch = BaseRecipeBranch("base_url", "{debupstream:tehname}", 0.4)
+        base_branch.substitute_branch_vars("tehname", wt.branch, revid2)
+        self.assertEqual("0.1", base_branch.deb_version)
+
+    def test_substitute_branch_vars_debupstream_pre_0_4(self):
+        wt = self.make_branch_and_tree("br")
+        cl_contents = ("package (0.1-1) unstable; urgency=low\n  * foo\n"
+                    " -- maint <maint@maint.org>  Tue, 04 Aug 2009 "
+                    "10:03:10 +0100\n")
+        self.build_tree_contents(
+            [("br/debian/", ), ('br/debian/changelog', cl_contents)])
+        wt.add(['debian', 'debian/changelog'])
+        revid = wt.commit("with changelog")
+        # In recipe format < 0.4 {debupstream} gets replaced from the resulting
+        # tree, not from the branch vars.
+        base_branch = BaseRecipeBranch("base_url", "{debupstream}", 0.2)
+        base_branch.substitute_branch_vars(None, wt.branch, revid)
+        self.assertEqual("{debupstream}", base_branch.deb_version)
 
     def test_list_branch_names(self):
         base_branch = BaseRecipeBranch("base_url", "1", 0.2)
@@ -1172,3 +1401,103 @@ class RecipeBranchTests(TestCaseInTempDir):
         self.assertEqual(
             ["merged", "nested", "merged_into_nested", "another_nested"],
             base_branch.list_branch_names())
+
+    def test_iter_all_branches(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.2)
+        merged_branch = RecipeBranch("merged", "merged_url")
+        base_branch.merge_branch(merged_branch)
+        nested_branch = RecipeBranch("nested", "nested_url")
+        merge_into_nested_branch = RecipeBranch("merged_into_nested", "another_url")
+        nested_branch.merge_branch(merge_into_nested_branch)
+        base_branch.nest_branch("subdir", nested_branch)
+        another_nested_branch = RecipeBranch("another_nested", "yet_another_url")
+        base_branch.merge_branch(another_nested_branch)
+        base_branch.run_command("a command")
+        self.assertEqual([
+            base_branch, merged_branch, nested_branch, merge_into_nested_branch,
+            another_nested_branch],
+            list(base_branch.iter_all_branches()))
+
+    def test_iter_all_instructions(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.2)
+        nested_branch = RecipeBranch("nested", "nested_url")
+        merge_into_nested_branch = RecipeBranch("merged_into_nested", "another_url")
+        nested_branch.merge_branch(merge_into_nested_branch)
+        merge_into_nested = nested_branch.child_branches[-1]
+        base_branch.run_command("a command")
+        cmd = base_branch.child_branches[-1]
+        base_branch.nest_branch("subdir", nested_branch)
+        nest = base_branch.child_branches[-1]
+        self.assertEqual([
+            cmd, nest, merge_into_nested],
+            list(base_branch.iter_all_instructions()))
+
+
+class DebUpstreamVariableTests(TestCase):
+
+    def write_changelog(self, version):
+        contents = textwrap.dedent("""
+            package (%s) experimental; urgency=low
+
+              * Initial release. (Closes: #XXXXXX)
+
+             -- Jelmer Vernooij <jelmer@debian.org>  Thu, 19 May 2011 10:07:41 +0100
+            """ % version)[1:]
+        return changelog.Changelog(file=contents)
+
+    def test_empty_changelog(self):
+        var = DebUpstreamVariable.from_changelog(None, changelog.Changelog())
+        self.assertRaises(SubstitutionUnavailable, var.get)
+
+    def test_version(self):
+        var = DebUpstreamVariable.from_changelog(None,
+            self.write_changelog("2.3"))
+        self.assertEquals("2.3", var.get())
+
+    def test_epoch(self):
+        # The epoch is (currently) ignored by {debupstream}.
+        var = DebUpstreamVariable.from_changelog(None,
+            self.write_changelog("2:2.3"))
+        self.assertEquals("2.3", var.get())
+
+    def test_base_without_snapshot(self):
+        var = DebUpstreamBaseVariable.from_changelog(None,
+            self.write_changelog("2.4"))
+        self.assertEquals("2.4+", var.get())
+
+    def test_base_with_svn_snapshot(self):
+        var = DebUpstreamBaseVariable.from_changelog(None,
+            self.write_changelog("2.4~svn4"))
+        self.assertEquals("2.4~", var.get())
+
+    def test_base_with_bzr_snapshot(self):
+        var = DebUpstreamBaseVariable.from_changelog(None,
+            self.write_changelog("2.4+bzr343"))
+        self.assertEquals("2.4+", var.get())
+
+
+class DebVersionVariableTests(TestCase):
+
+    def write_changelog(self, version):
+        contents = textwrap.dedent("""
+            package (%s) experimental; urgency=low
+
+              * Initial release. (Closes: #XXXXXX)
+
+             -- Jelmer Vernooij <jelmer@debian.org>  Thu, 19 May 2011 10:07:41 +0100
+            """ % version)[1:]
+        return changelog.Changelog(file=contents)
+
+    def test_empty_changelog(self):
+        var = DebVersionVariable.from_changelog(None, changelog.Changelog())
+        self.assertRaises(SubstitutionUnavailable, var.get)
+
+    def test_simple(self):
+        var = DebVersionVariable.from_changelog(
+            None, self.write_changelog("2.3-1"))
+        self.assertEquals("2.3-1", var.get())
+
+    def test_epoch(self):
+        var = DebVersionVariable.from_changelog(
+            None, self.write_changelog("4:2.3-1"))
+        self.assertEquals("4:2.3-1", var.get())
